from svvamp import RuleIRVDuels
from svvamp import Profile


def test_loser_equals_selected_two():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [1, 0, 1],
        ... ])
        >>> rule = RuleIRVDuels()(profile)
        >>> rule.w_
        0
    """
    pass



# -*- coding: utf-8 -*-
"""
Created on 7 dec. 2018, 15:01
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.rules.rule_irv import RuleIRV
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile
from svvamp.utils.pseudo_bool import equal_true
from svvamp.utils.misc import preferences_ut_to_matrix_duels_ut


class RuleICRV(Rule):
    """Instant-Condorcet Runoff Voting (ICRV).

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleICRV()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [[2. 0. 1.]]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [[2. 1. 0.]]
        w = 0
        score_w = [2.]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (True)      ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (True)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = lazy
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = lazy
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = fast, um_option = lazy, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 1. 2.]
        sufficient_coalition_size_cm =
        [0. 6. 4.]

    Notes
    -----
    Principle: eliminate candidates as in IRV; stop as soon as there is a Condorcet winner.

    * Even round ``r`` (including round 0): if a candidate ``w`` has only victories against all other non-eliminated
      candidates (i.e. is a Condorcet winner in this subset, in the sense of :attr:`matrix_victories_rk`), then ``w``
      is declared the winner.
    * Odd round ``r``: the candidate who is ranked first (among non-eliminated candidates) by least voters is
    eliminated, like in :class:`RuleIRV`.

    This method meets the Condorcet criterion.

    * :meth:`is_cm_`:

        * :attr:`cm_option` = ``'fast'``: Rely on :class:`RuleIRV`'s fast algorithm. Polynomial heuristic. Can prove
          CM but unable to decide non-CM (except in rare obvious cases).
        * :attr:`cm_option` = ``'slow'``: Rely on :class:`RuleExhaustiveBallot`'s exact algorithm. Non-polynomial
          heuristic (:math:`2^{n_c}`). Quite efficient to prove CM or non-CM.
        * :attr:`cm_option` = ``'very_slow'``: Rely on :class:`RuleIRV`'s exact algorithm. Non-polynomial
          heuristic (:math:`n_c!`). Very efficient to prove CM or non-CM.
        * :attr:`cm_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

        Each algorithm above exploits the faster ones. For example, if :attr:`cm_option` = ``'very_slow'``,
        SVVAMP tries the fast algorithm first, then the slow one, then the 'very slow' one. As soon as it reaches
        a decision, computation stops.

    * :meth:`is_icm_`: Exact in polynomial time.
    * :meth:`is_im_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`~svvamp.Election.not_iia`: Exact in polynomial time.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.

    References
    ----------
    'Four Condorcet-Hare Hybrid Methods for Single-Winner Elections', James Green-Armytage, 2011.

    See Also
    --------
    :class:`RuleExhaustiveBallot`, :class:`RuleIRV`, :class:`RuleIRVDuels`, :class:`RuleCondorcetAbsIRV`,
    :class:`RuleCondorcetVtbIRV`.
    """

    full_name = 'Benham'
    abbreviation = 'Ben'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'cm_option': {'allowed': {'fast', 'slow', 'very_slow', 'exact'}, 'default': 'fast'},
        'icm_option': {'allowed': ['exact'], 'default': 'exact'},
    })

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=False,
            log_identity="ICRV", **kwargs
        )

    def __call__(self, profile):
        self.delete_cache(suffix='_')
        self.profile_ = profile
        # Grab the IRV ballot of the profile (or create it)
        irv_options = {}
        if self.cm_option == 'fast':
            irv_options['cm_option'] = 'fast'
        elif self.cm_option == 'slow':
            irv_options['cm_option'] = 'slow'
        else:  # self.cm_option in {'very_slow', 'exact'}:
            irv_options['cm_option'] = 'exact'
        self.irv_ = RuleIRV(**irv_options)(self.profile_)
        return self

    # %% Counting the ballots

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        scores = []
        is_candidate_alive = np.ones(self.profile_.n_c, dtype=np.bool)
        worst_to_best = []
        r = 0
        while True:
            # Here, scores_r is the number of victories (when restricting to alive candidates).
            scores_r = np.full(self.profile_.n_c, np.nan)
            scores_r[is_candidate_alive] = np.sum(
                self.profile_.matrix_victories_rk[is_candidate_alive, :][:, is_candidate_alive], 1)
            scores.append(scores_r)
            best_c = np.nanargmax(scores_r)
            self.mylogv("best_c =", best_c, 3)
            if scores_r[best_c] == self.profile_.n_c - 1 - r:
                w = best_c
                self.mylogm("scores (before) =", scores, 3)
                scores = np.array(scores)
                self.mylogm("scores (after) =", scores, 3)
                candidates_alive = np.array(range(self.profile_.n_c))[is_candidate_alive]
                candidates_by_scores_best_to_worst = np.concatenate((
                    candidates_alive[np.argsort(-scores_r[is_candidate_alive], kind='mergesort')],
                    worst_to_best[::-1]
                )).astype(dtype=np.int)
                return {'scores': scores, 'w': w,
                        'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst}

            # Now, scores_r is the Plurality score (IRV-style)
            scores_r = np.full(self.profile_.n_c, np.nan)
            scores_r[is_candidate_alive] = np.sum(
                self.profile_.preferences_borda_rk[:, is_candidate_alive]
                == np.max(self.profile_.preferences_borda_rk[:, is_candidate_alive], 1)[:, np.newaxis],
                0)
            scores.append(scores_r)
            loser = np.where(scores_r == np.nanmin(scores_r))[0][-1]
            self.mylogv("loser =", loser, 3)
            is_candidate_alive[loser] = False
            worst_to_best.append(loser)

            r += 1

    @cached_property
    def w_(self):
        self.mylog("Compute w", 1)
        is_alive = np.ones(self.profile_.n_c, dtype=np.bool)
        nb_alive = self.profile_.n_c
        preferences_borda_rk = self.profile_.preferences_borda_rk.copy()  # We will put -1 when a candidate loses
        scores_r = self.profile_.plurality_scores_rk.copy().astype(float)
        ballots_r = self.profile_.preferences_rk[:, 0].copy()
        loser = None
        for r in range(self.profile_.n_c - 1):
            # Is there a Condorcet winner?
            is_condorcet_winner = (
                is_alive
                & (np.sum(self.profile_.matrix_victories_rk[:, is_alive], axis=1) == nb_alive - 1)
            )
            condorcet_winners = np.where(is_condorcet_winner)[0]
            if condorcet_winners.size > 0:
                return condorcet_winners[0]
            # Now, scores_r is the Plurality score (IRV-style)
            if r != 0:
                preferences_borda_rk[:, loser] = -1
                new_ballots = np.argmax(preferences_borda_rk[ballots_r == loser, :], axis=1)
                ballots_r[ballots_r == loser] = new_ballots
                scores_r += np.bincount(new_ballots, minlength=self.profile_.n_c)
                scores_r[loser] = np.nan
            # Who gets eliminated?
            loser = np.where(scores_r == np.nanmin(scores_r))[0][-1]  # Tie-breaking: the last index
            is_alive[loser] = False
            nb_alive -= 1
            self.mylogv("loser =", loser, 3)
            # Save time if last round
            if nb_alive == 1:
                return np.where(is_alive)[0][0]

    @cached_property
    def scores_(self):
        """2d array.

        For even rounds ``r`` (including round 0), ``scores[r, c]`` is the number of victories for ``c`` in
        :attr:`matrix_victories_rk` (restricted to non-eliminated candidates). Ties count for 0.5.

        For odd rounds ``r``, ``scores[r, c]`` is the number of voters who rank ``c`` first (among non-eliminated
        candidates).
        """
        return self._count_ballots_['scores']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers.

        Candidates that are not eliminated at the moment a Condorcet winner is detected are sorted by their number of
        victories in :attr:`matrix_victories_rk` (restricted to candidates that are not eliminated at that time).

        Other candidates are sorted in the reverse order of their IRV elimination.
        """
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    # TODO: self._v_might_im_for_c

    @cached_property
    def meets_majority_favorite_c_rk_ctb(self):
        return True

    @cached_property
    def meets_condorcet_c_rk(self):
        return True

    # %% Coalition Manipulation (CM)

    @property
    def losing_candidates_(self):
        """If ``irv_.w_ does not win, then we put her first. Other losers are sorted as usual. (scores in
        ``matrix_duels_ut``).
        """
        self.mylog("Compute ordered list of losing candidates", 1)
        if self.w_ == self.irv_.w_:
            # As usual
            losing_candidates = np.concatenate((
                np.array(range(0, self.w_), dtype=int), np.array(range(self.w_ + 1, self.profile_.n_c), dtype=int)))
            losing_candidates = losing_candidates[np.argsort(
                - self.profile_.matrix_duels_ut[losing_candidates, self.w_], kind='mergesort')]
        else:
            # Put irv_.w_ first.
            losing_candidates = np.array(
                [c for c in range(self.profile_.n_c) if c != self.w_ and c != self.irv_.w_]).astype(np.int)
            losing_candidates = losing_candidates[np.argsort(
                - self.profile_.matrix_duels_ut[losing_candidates, self.w_], kind='mergesort')]
            losing_candidates = np.concatenate(([self.irv_.w_], losing_candidates))
        return losing_candidates

    def _cm_aux_(self, c, ballots_m, preferences_rk_s):
        profile_test = Profile(preferences_rk=np.concatenate((preferences_rk_s, ballots_m)), sort_voters=False)
        if profile_test.n_v != self.profile_.n_v:
            raise AssertionError('Uh-oh!')
        winner_test = self.__class__()(profile_test).w_
        return winner_test == c

    def _cm_main_work_c_(self, c, optimize_bounds):
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        n_s = self.profile_.n_v - n_m
        candidates = np.array(range(self.profile_.n_c))
        preferences_borda_s = self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        preferences_rk_s = self.profile_.preferences_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        matrix_duels_vtb_temp = (preferences_ut_to_matrix_duels_ut(preferences_borda_s))
        self.mylogm("CM: matrix_duels_vtb_temp =", matrix_duels_vtb_temp, 3)
        # More preliminary checks. It's more convenient to put them in that method, because we need
        # ``preferences_borda_s`` and ``matrix_duels_vtb_temp``.
        d_neq_c = (np.array(range(self.profile_.n_c)) != c)
        # Prevent another cond. Look at the weakest duel for ``d``, she has ``matrix_duels_vtb_temp[d, e]``. We simply
        # need that:
        # ``matrix_duels_vtb_temp[d, e] <= (n_s + n_m) / 2``
        # ``2 * max_d(min_e(matrix_duels_vtb_temp[d, e])) - n_s <= n_m``
        n_manip_prevent_cond = 0
        for d in candidates[d_neq_c]:
            e_neq_d = (np.array(range(self.profile_.n_c)) != d)
            n_prevent_d = np.maximum(2 * np.min(matrix_duels_vtb_temp[d, e_neq_d]) - n_s, 0)
            n_manip_prevent_cond = max(n_manip_prevent_cond, n_prevent_d)
        self.mylogv("CM: n_manip_prevent_cond =", n_manip_prevent_cond, 3)
        self._update_necessary(self._necessary_coalition_size_cm, c, n_manip_prevent_cond,
                               'CM: Update necessary_coalition_size_cm[c] = n_manip_prevent_cond =')
        if not optimize_bounds and self._necessary_coalition_size_cm[c] > n_m:
            return True

        # Let us work
        if self.w_ == self.irv_.w_:
            self.mylog('CM: c != self.irv_.w == self.w', 3)
            if self.cm_option == "fast":
                self.irv_.cm_option = "fast"
            elif self.cm_option == "slow":
                self.irv_.cm_option = "slow"
            else:
                self.irv_.cm_option = "exact"
            irv_is_cm_c = self.irv_.is_cm_c_(c)
            if equal_true(irv_is_cm_c):
                # Use IRV without bounds
                self.mylog('CM: Use IRV without bounds')
                suggested_path_one = self.irv_.example_path_cm_c_(c)
                self.mylogv("CM: suggested_path =", suggested_path_one, 3)
                ballots_m = self.irv_.example_ballots_cm_c_(c)
                manipulation_found = self._cm_aux_(c, ballots_m, preferences_rk_s)
                self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                if manipulation_found:
                    self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                            'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                    # We will not do better with any algorithm (even the brute force algo).
                    return False
                # Use IRV with bounds
                self.irv_.is_cm_c_with_bounds_(c)
                self.mylog('CM: Use IRV with bounds')
                suggested_path_two = self.irv_.example_path_cm_c_(c)
                self.mylogv("CM: suggested_path =", suggested_path_two, 3)
                if np.array_equal(suggested_path_one, suggested_path_two):
                    self.mylog('CM: Same suggested path as before, skip computation')
                else:
                    ballots_m = self.irv_.example_ballots_cm_c_(c)
                    manipulation_found = self._cm_aux_(c, ballots_m, preferences_rk_s)
                    self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                    if manipulation_found:
                        self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                                'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                        # We will not do better with any algorithm (even the brute force algo).
                        return False
        else:  # self.w_ != self.irv_.w_:
            if c == self.irv_.w_:
                self.mylog('CM: c == self.irv_.w != self._w', 3)
                suggested_path = self.irv_.elimination_path_
                self.mylogv("CM: suggested_path =", suggested_path, 3)
                ballots_m = self.irv_.example_ballots_cm_w_against_(w_other_rule=self.w_)
                manipulation_found = self._cm_aux_(c, ballots_m, preferences_rk_s)
                self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                if manipulation_found:
                    self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                            'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                    # We will not do better with any algorithm (even the brute force algo).
                    return
            else:
                self.mylog('CM: c, self.irv_.w_ and self.w_ are all distinct', 3)
        if self.cm_option == 'exact':
            return self._cm_main_work_c_exact_(c, optimize_bounds)



# -*- coding: utf-8 -*-
"""
Created on nov. 29, 2018, 09:56
Copyright François Durand 2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
from svvamp.preferences.generator_profile_noise import GeneratorProfileNoise
from svvamp.preferences.profile_from_file import ProfileFromFile


class GeneratorProfileNoisedFile(GeneratorProfileNoise):
    """Profile generator loading a file, then adding noise

    Parameters
    ----------
    file_name : str
        The file for the initial profile.
    relative_noise : number
        The relative noise.
    absolute_noise : number
        The absolute noise
    sort_voters : bool
        This argument is passed to :class:`Profile`.

    Notes
    -----
    This class is just a combination of :class:`ProfileFromFile` and :class:`GeneratorProfileNoise`.
    """

    def __init__(self, file_name, relative_noise=0., absolute_noise=0., sort_voters=False):
        self.file_name = file_name
        base_profile = ProfileFromFile(file_name=file_name, sort_voters=False)
        super().__init__(base_profile=base_profile, relative_noise=relative_noise, absolute_noise=absolute_noise,
                         sort_voters=sort_voters)
        self.log_creation = ['Noised File', 'File name', file_name,
                             'Relative noise', relative_noise, 'Absolute noise', absolute_noise]



# -*- coding: utf-8 -*-
"""
Created on nov. 03, 2014, 23:33
Copyright François Durand 2014, 2015
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""

import time
import numpy as np
import pandas as pd

from svvamp.meta.study_profile_criteria import StudyProfileCriteria
from svvamp.meta.voting_rule_tasks import VotingRuleTasks
from svvamp.preferences.generator_profile_noise import GeneratorProfileNoise
from svvamp.utils.pseudo_bool import equal_true


class ExperimentAnalyzer:

    def __init__(self,
                 base_profile=None,
                 n_samples=1,
                 relative_noise=0.,
                 absolute_noise=0.,
                 study_profile_criteria=None,
                 voting_rule_tasks=None,
                 output_dir='out',
                 output_file_suffix='',
                 log_csv=None,
                 ping_period=10):
        # Check sanity (or not)
        if study_profile_criteria:
            study_profile_criteria.check_sanity()
        if voting_rule_tasks:
            voting_rule_tasks.check_sanity()
        # Default options
        if study_profile_criteria is None:
            study_profile_criteria = StudyProfileCriteria()
        if voting_rule_tasks is None:
            voting_rule_tasks = VotingRuleTasks()
        if log_csv is None:
            log_csv = []
        # Set the default attributes
        self.default_base_profile = base_profile
        self.default_n_samples = n_samples
        self.default_relative_noise = relative_noise
        self.default_absolute_noise = absolute_noise
        self.default_study_profile_criteria = study_profile_criteria
        self.default_voting_rule_tasks = voting_rule_tasks
        self.default_output_dir = output_dir
        self.default_output_file_suffix = output_file_suffix
        self.default_log_csv = log_csv
        self.default_ping_period = ping_period
        # Declare the attributes
        self.base_profile = None
        self.n_samples = None
        self.relative_noise = None
        self.absolute_noise = None
        self.study_profile_criteria = None
        self.voting_rule_tasks = None
        self.output_dir = None
        self.output_file_suffix = None
        self.log_csv = None
        self.ping_period = None

    def check_sanity(self):
        self.study_profile_criteria.check_sanity()
        self.voting_rule_tasks.check_sanity()

    def _prepare_csv(self):
        prefix = self.output_dir + '/Results_' + time.strftime('%Y-%m-%d_%H-%M-%S')
        suffix = '_' + self.output_file_suffix if self.output_file_suffix else ''
        self.file_d_en = open(f"{prefix}_d_en{suffix}.csv", 'a')
        self.file_d_fr = open(f"{prefix}_d_fr{suffix}.csv", 'a')
        self.file_m_en = open(f"{prefix}_m_en{suffix}.csv", 'a')
        self.file_m_fr = open(f"{prefix}_m_fr{suffix}.csv", 'a')
        headers_m = pd.DataFrame([
            'Rule', 'Rule (abbr)', 'Rule (class)',
            'Criterion', 'Candidate', 'Candidate name',
            'Algorithm parameters',
            'Computation time',
            'Rate (lower bound)',
            'Rate (upper bound)',
            'Rate (uncertainty)',
            'Number of profiles',
            'V', 'C', 'Culture type',
            'Culture parameter 1', 'Culture parameter 1 value',
            'Culture parameter 2', 'Culture parameter 2 value',
            'Culture parameter 3', 'Culture parameter 3 value',
            'Culture parameter 4', 'Culture parameter 4 value'
        ]).T
        headers_m.to_csv(path_or_buf=self.file_m_en, index=False, header=False, line_terminator='\n')
        headers_m.to_csv(path_or_buf=self.file_m_fr, index=False, header=False, sep=';', line_terminator='\n')

    def _df_to_csv_m(self, df):
        return _df_to_csv(df, file_en=self.file_m_en, file_fr=self.file_m_fr)

    def _df_to_csv_d(self, df):
        return _df_to_csv(df, file_en=self.file_d_en, file_fr=self.file_d_fr)

    def __call__(
        self,
        base_profile=None,
        n_samples=None,
        relative_noise=None,
        absolute_noise=None,
        study_profile_criteria=None,
        voting_rule_tasks=None,
        output_dir=None,
        output_file_suffix=None,
        log_csv=None,
        ping_period=None
    ):
        # Set the parameters
        self.base_profile = self.default_base_profile if base_profile is None else base_profile
        self.n_samples = self.default_n_samples if n_samples is None else n_samples
        self.relative_noise = self.default_relative_noise if relative_noise is None else relative_noise
        self.absolute_noise = self.default_absolute_noise if absolute_noise is None else absolute_noise
        self.study_profile_criteria = (self.default_study_profile_criteria if study_profile_criteria is None
                                       else study_profile_criteria)
        self.voting_rule_tasks = self.default_voting_rule_tasks if voting_rule_tasks is None else voting_rule_tasks
        self.output_dir = self.default_output_dir if output_dir is None else output_dir
        self.output_file_suffix = self.default_output_file_suffix if output_file_suffix is None else output_file_suffix
        self.log_csv = self.default_log_csv if log_csv is None else log_csv
        self.ping_period = self.default_ping_period if ping_period is None else ping_period
        # Check sanity (or not)
        if study_profile_criteria:
            study_profile_criteria.check_sanity()
        if voting_rule_tasks:
            voting_rule_tasks.check_sanity()
        # Initialize
        n_c = self.base_profile.n_c
        n_v = self.base_profile.n_v
        labels_candidates = self.base_profile.labels_candidates
        # - Initialize records for profile results
        results_p_boolean_criteria = [dict(inf=0, sup=0) for _ in self.study_profile_criteria.boolean_criteria]
        results_p_numerical_criteria = [[] for _ in self.study_profile_criteria.numerical_criteria]
        results_p_special_candidates_criteria = [
            np.zeros(n_c + 1) for _ in self.study_profile_criteria.special_candidates_criteria
        ]  # The last coefficient will be used for cases np.nan
        results_p_array_criteria = [np.zeros(n_c) for _ in self.study_profile_criteria.array_criteria]
        results_p_matrix_criteria = [
            np.zeros(getattr(self.base_profile, criterion).shape)
            for criterion in self.study_profile_criteria.matrix_criteria
        ]
        # - Initialize record for voting systems results
        results_vs_winner = [np.zeros(n_c) for _ in self.voting_rule_tasks]
        results_vs_result_criteria = [
            [
                dict(inf=0, sup=0)
                for _ in study_rule_criteria.result_criteria
            ]
            for (rule_class, options, study_rule_criteria) in self.voting_rule_tasks
        ]
        results_vs_numerical_criteria = [
            [
                dict(inf=0, sup=0, time=0)
                for _ in study_rule_criteria.numerical_criteria
            ]
            for (rule_class, options, study_rule_criteria) in self.voting_rule_tasks
        ]
        results_vs_manipulation_criteria = [
            [
                dict(inf=0, sup=0, log='', time=0)
                for _ in study_rule_criteria.manipulation_criteria
            ]
            for (rule_class, options, study_rule_criteria) in self.voting_rule_tasks
        ]
        results_vs_manipulation_criteria_c = [
            [
                dict(inf=np.zeros(n_c), sup=np.zeros(n_c), log='', time=0)
                for _ in study_rule_criteria.manipulation_criteria_c
            ]
            for (rule_class, options, study_rule_criteria) in self.voting_rule_tasks
        ]
        results_vs_utility_criteria = [
            [
                []
                for _ in study_rule_criteria.utility_criteria
            ]
            for (rule_class, options, study_rule_criteria) in self.voting_rule_tasks
        ]
        self._prepare_csv()
        generator_profile = GeneratorProfileNoise(
            base_profile=self.base_profile, relative_noise=self.relative_noise, absolute_noise=self.absolute_noise)
        # Heavy lifting
        print('Compute...')
        for n_iterations in range(self.n_samples):
            profile = generator_profile()
            # Tests on the profile itself
            for i_criterion, criterion in enumerate(self.study_profile_criteria.boolean_criteria):
                answer = getattr(profile, criterion)
                if equal_true(answer):
                    results_p_boolean_criteria[i_criterion]['inf'] += 1
                    results_p_boolean_criteria[i_criterion]['sup'] += 1
                elif np.isnan(answer):
                    results_p_boolean_criteria[i_criterion]['sup'] += 1
            for i_criterion, (criterion, func, name) in enumerate(self.study_profile_criteria.numerical_criteria):
                answer = getattr(profile, criterion)
                results_p_numerical_criteria[i_criterion].append(answer)
            for i_criterion, criterion in enumerate(self.study_profile_criteria.special_candidates_criteria):
                answer = getattr(profile, criterion)
                if np.isnan(answer):
                    results_p_special_candidates_criteria[i_criterion][-1] += 1
                else:
                    results_p_special_candidates_criteria[i_criterion][answer] += 1
            for i_criterion, criterion in enumerate(self.study_profile_criteria.array_criteria):
                answer = getattr(profile, criterion)
                results_p_array_criteria[i_criterion] += answer
            for i_criterion, criterion in enumerate(self.study_profile_criteria.matrix_criteria):
                answer = getattr(profile, criterion)
                results_p_matrix_criteria[i_criterion] += answer
            # Tests on the voting systems
            for i_task, (rule_class, options, study_rule_criteria) in enumerate(self.voting_rule_tasks):
                election = rule_class(**options)(profile)
                # The winner
                results_vs_winner[i_task][election.w_] += 1
                # Result criteria
                for i_criterion, criterion in enumerate(study_rule_criteria.result_criteria):
                    answer = getattr(election, criterion)
                    if equal_true(answer):
                        results_vs_result_criteria[i_task][i_criterion]['inf'] += 1
                        results_vs_result_criteria[i_task][i_criterion]['sup'] += 1
                    elif np.isnan(answer):
                        results_vs_result_criteria[i_task][i_criterion]['sup'] += 1
                # Utility criteria
                for i_criterion, (criterion, func, name) in enumerate(study_rule_criteria.utility_criteria):
                    answer = getattr(election, criterion)
                    results_vs_utility_criteria[i_task][i_criterion].append(answer)
                # Numerical criteria
                for i_criterion, criterion in enumerate(study_rule_criteria.numerical_criteria):
                    t1 = time.time()
                    inf, sup = getattr(election, criterion)
                    t2 = time.time()
                    results_vs_numerical_criteria[i_task][i_criterion]['time'] += t2 - t1
                    results_vs_numerical_criteria[i_task][i_criterion]['inf'] = inf
                    results_vs_numerical_criteria[i_task][i_criterion]['sup'] = sup
                # Manipulation criteria (by candidate)
                for i_criterion, criterion in enumerate(study_rule_criteria.manipulation_criteria_c):
                    t1 = time.time()
                    answer = getattr(election, criterion)
                    t2 = time.time()
                    results_vs_manipulation_criteria_c[i_task][i_criterion]['time'] += t2 - t1
                    results_vs_manipulation_criteria_c[i_task][i_criterion]['inf'] += np.equal(answer, True)
                    results_vs_manipulation_criteria_c[i_task][i_criterion]['sup'] += np.equal(answer, True)
                    results_vs_manipulation_criteria_c[i_task][i_criterion]['sup'] += np.isnan(answer)
                    if n_iterations == 0:
                        log = election.log_(criterion)
                        results_vs_manipulation_criteria_c[i_task][i_criterion]['log'] += log
                # Manipulation criteria (Boolean)
                for i_criterion, criterion in enumerate(study_rule_criteria.manipulation_criteria):
                    t1 = time.time()
                    answer = getattr(election, criterion)
                    t2 = time.time()
                    results_vs_manipulation_criteria[i_task][i_criterion]['time'] += t2 - t1
                    if equal_true(answer):
                        results_vs_manipulation_criteria[i_task][i_criterion]['inf'] += 1
                        results_vs_manipulation_criteria[i_task][i_criterion]['sup'] += 1
                    elif np.isnan(answer):
                        results_vs_manipulation_criteria[i_task][i_criterion]['sup'] += 1
                    if n_iterations == 0:
                        log = election.log_(criterion)
                        results_vs_manipulation_criteria[i_task][i_criterion]['log'] += log
            # Report progression (or not)
            if (n_iterations + 1) % self.ping_period == 0:
                print(f'{n_iterations + 1} profiles analyzed')

        # Write results in csv
        print('Write results in csv files...')
        # Results on profiles
        for i_criterion, criterion in enumerate(self.study_profile_criteria.boolean_criteria):
            lower = results_p_boolean_criteria[i_criterion]['inf'] / self.n_samples
            upper = results_p_boolean_criteria[i_criterion]['sup'] / self.n_samples
            self._df_to_csv_m(
                ['Profile', '', '', criterion, '', '', '', '', lower, upper, upper - lower, self.n_samples, n_v, n_c]
                + self.log_csv)
        for i_criterion, (criterion, func, name) in enumerate(self.study_profile_criteria.numerical_criteria):
            answer = func(results_p_numerical_criteria[i_criterion])
            self._df_to_csv_m(['Profile', '', '', name, '', '', '', '', answer, answer, 0, self.n_samples, n_v, n_c]
                              + self.log_csv)
        for i_criterion, criterion in enumerate(self.study_profile_criteria.special_candidates_criteria):
            rate = results_p_special_candidates_criteria[i_criterion] / self.n_samples
            for c in range(n_c):
                self._df_to_csv_m(
                    ['Profile', '', '', criterion, c, labels_candidates[c], '', '', rate[c], rate[c], 0, self.n_samples,
                     n_v, n_c]
                    + self.log_csv
                )
            self._df_to_csv_m(
                ['Profile', '', '', criterion, 'None', 'None', '', '', rate[n_c], rate[n_c], 0, self.n_samples,
                 n_v, n_c]
                + self.log_csv
            )
        for i_criterion, criterion in enumerate(self.study_profile_criteria.array_criteria):
            aver = results_p_array_criteria[i_criterion] / self.n_samples
            for c in range(n_c):
                self._df_to_csv_m(
                    ['Profile', '', '', criterion, c, labels_candidates[c], '', '', aver[c], aver[c], 0, self.n_samples,
                     n_v, n_c]
                    + self.log_csv
                )
        for i_criterion, criterion in enumerate(self.study_profile_criteria.matrix_criteria):
            self._df_to_csv_d([criterion])
            self._df_to_csv_d(results_p_matrix_criteria[i_criterion] / self.n_samples)
        # Results on voting systems
        for i_task, (rule_class, options, study_rule_criteria) in enumerate(self.voting_rule_tasks):
            # Winner
            rate_win = results_vs_winner[i_task] / self.n_samples
            for c in range(n_c):
                self._df_to_csv_m(
                    [
                        rule_class.full_name, rule_class.abbreviation, rule_class.__name__,
                        'Winner', c, labels_candidates[c], '', '',
                        rate_win[c], rate_win[c], 0, self.n_samples, n_v, n_c
                    ]
                    + self.log_csv
                )
            # Result criteria
            for i_criterion, criterion in enumerate(study_rule_criteria.result_criteria):
                lower = results_vs_result_criteria[i_task][i_criterion]['inf'] / self.n_samples
                upper = results_vs_result_criteria[i_task][i_criterion]['sup'] / self.n_samples
                self._df_to_csv_m(
                    [rule_class.full_name, rule_class.abbreviation, rule_class.__name__,
                     criterion, '', '', '', '', lower, upper, upper - lower, self.n_samples, n_v, n_c]
                    + self.log_csv
                )
            # Manipulation criteria
            for i_criterion, criterion in enumerate(study_rule_criteria.manipulation_criteria):
                lower = results_vs_manipulation_criteria[i_task][i_criterion]['inf'] / self.n_samples
                upper = results_vs_manipulation_criteria[i_task][i_criterion]['sup'] / self.n_samples
                the_time = results_vs_manipulation_criteria[i_task][i_criterion]['time']
                the_log = results_vs_manipulation_criteria[i_task][i_criterion]['log']
                self._df_to_csv_m(
                    [
                        rule_class.full_name, rule_class.abbreviation, rule_class.__name__,
                        criterion, 'any', '', the_log, the_time, lower, upper, upper - lower,
                        self.n_samples, n_v, n_c
                    ]
                    + self.log_csv
                )
            for i_criterion, criterion in enumerate(study_rule_criteria.manipulation_criteria_c):
                lower = results_vs_manipulation_criteria_c[i_task][i_criterion]['inf'] / self.n_samples
                upper = results_vs_manipulation_criteria_c[i_task][i_criterion]['sup'] / self.n_samples
                the_time = results_vs_manipulation_criteria_c[i_task][i_criterion]['time']
                the_log = results_vs_manipulation_criteria_c[i_task][i_criterion]['log']
                for c in range(n_c):
                    self._df_to_csv_m(
                        [
                            rule_class.full_name, rule_class.abbreviation, rule_class.__name__,
                            criterion, c, labels_candidates[c], the_log, '', lower[c], upper[c], upper[c] - lower[c],
                            self.n_samples, n_v, n_c
                        ]
                        + self.log_csv
                    )
                # Indicate the time only once
                self._df_to_csv_m(
                    [rule_class.full_name, rule_class.abbreviation, rule_class.__name__,
                     criterion, 'any', '', the_log, the_time, '', '', '', self.n_samples, n_v, n_c]
                    + self.log_csv
                )
            # Utility criteria
            for i_criterion, (criterion, func, name) in enumerate(study_rule_criteria.utility_criteria):
                answer = func(results_vs_utility_criteria[i_task][i_criterion])
                self._df_to_csv_m(
                    [rule_class.full_name, rule_class.abbreviation, rule_class.__name__,
                     name, '', '', '', '', answer, answer, 0, self.n_samples, n_v, n_c]
                    + self.log_csv
                )
            # Numerical criteria
            for i_criterion, criterion in enumerate(study_rule_criteria.numerical_criteria):
                lower = results_vs_numerical_criteria[i_task][i_criterion]['inf'] / self.n_samples
                upper = results_vs_numerical_criteria[i_task][i_criterion]['sup'] / self.n_samples
                the_time = results_vs_numerical_criteria[i_task][i_criterion]['time']
                try:
                    self._df_to_csv_m(
                        [
                            rule_class.full_name, rule_class.abbreviation, rule_class.__name__,
                            criterion, 'any', '', '', the_time, lower, upper, upper - lower,
                            self.n_samples, n_v, n_c
                        ]
                        + self.log_csv
                    )
                except ValueError:
                    print(criterion)
                    print(lower)
                    print(upper)
                    raise ValueError
        self.file_m_en.close()
        self.file_m_fr.close()
        self.file_d_en.close()
        self.file_d_fr.close()
        print('Simulation finished')


def my_float(x):
    if isinstance(x, float):
        return ('%.6f' % x).replace('.', ',')
    else:
        return x


def _df_to_csv(df, file_en, file_fr):
    df = pd.DataFrame(df)
    if df.shape[1] == 1:
        df = df.T
    df.to_csv(path_or_buf=file_en, index=False, header=False, line_terminator='\n')
    df.apply(my_float).to_csv(path_or_buf=file_fr, index=False, header=False, sep=';', line_terminator='\n')



# -*- coding: utf-8 -*-
"""
Created on oct. 30, 2014, 23:52
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.

==============

The function preflib_to_preferences_ut below is adapted from
PreflibUtils.py by Nicholas Mattei. We reproduce its license here.

File: 	PrefLibUtilities.py
Author:	Nicholas Mattei (nicholas.mattei@nicta.com.au)
Date:	April 4, 2013
        November 6th, 2013

  * Copyright (c) 2014, Nicholas Mattei and NICTA
  * All rights reserved.
  *
  * Developed by: Nicholas Mattei
  *               NICTA
  *               http://www.nickmattei.net
  *               http://www.preflib.org
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
  *     * Redistributions of source code must retain the above copyright
  *       notice, this list of conditions and the following disclaimer.
  *     * Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
  *     * Neither the name of NICTA nor the
  *       names of its contributors may be used to endorse or promote products
  *       derived from this software without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY NICTA ''AS IS'' AND ANY
  * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL NICTA BE LIABLE FOR ANY
  * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import numpy as np
from svvamp.utils.misc import preferences_ut_to_preferences_borda_ut


def preflib_to_preferences_ut(file_name):
    # PreflibUtils.py
    input_file = open(file_name, 'r')
    # Number of candidates
    the_line = input_file.readline()
    # Map of candidates
    n_c = int(the_line.strip())
    candidates_map = {}
    labels_candidates = []
    for c in range(n_c):
        bits = input_file.readline().strip().split(",")
        candidates_map[int(bits[0].strip()) - 1] = bits[1].strip()
        labels_candidates.append(bits[1].strip())
    # Now we have V, sum_of_vote_count, num_unique orders
    bits = input_file.readline().strip().split(",")
    n_v = int(bits[0].strip())
    unique_orders = int(bits[2].strip())
    # Now, the ballots themselves
    preferences_utilities = np.zeros((n_v, n_c))
    start_index = 0
    for i in range(unique_orders):
        rec = input_file.readline().strip()
        # need to parse the rec properly..
        count = int(rec[:rec.index(",")])
        bits = rec[rec.index(",")+1:].strip().split(",")
        ballot_temp = np.full(n_c, - n_c)
        if rec.find("{") == -1:
            # its strict, just split on ,
            for crank in range(len(bits)):
                ballot_temp[int(bits[crank]) - 1] = - crank
        else:
            crank = 0
            partial = False
            for ccand in bits:
                if ccand.find("{") != -1:
                    partial = True
                    t = ccand.replace("{", "")
                    ballot_temp[int(t.strip()) - 1] = - crank
                elif ccand.find("}") != -1:
                    partial = False
                    t = ccand.replace("}", "")
                    ballot_temp[int(t.strip()) - 1] = - crank
                    crank += 1
                else:
                    ballot_temp[int(ccand.strip()) - 1] = - crank
                    if not partial:
                        crank += 1
        preferences_utilities[start_index:start_index + count, :] = (
            preferences_ut_to_preferences_borda_ut(ballot_temp[np.newaxis, :]) - (n_c - 1) / 2)
        start_index += count
    input_file.close()
    return preferences_utilities, labels_candidates



from svvamp import RuleCoombs, Profile


def test_um_option_setter():
    """
        >>> rule = RuleCoombs()
        >>> rule.um_option = 'fast'
        >>> rule.um_option
        'fast'
        >>> rule.um_option = 'exact'
        >>> rule.um_option
        'exact'
        >>> rule.um_option = 'foo'
        Traceback (most recent call last):
        ValueError: Unknown value for um_option: foo
    """
    pass


def test_cm_option_setter():
    """
        >>> rule = RuleCoombs()
        >>> rule.cm_option = 'fast'
        >>> rule.cm_option
        'fast'
        >>> rule.cm_option = 'exact'
        >>> rule.cm_option
        'exact'
        >>> rule.cm_option = 'foo'
        Traceback (most recent call last):
        ValueError: Unknown value for cm_option: foo
    """
    pass


def test_is_um_c():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [2, 0, 1],
        ... ])
        >>> rule = RuleCoombs()(profile)
        >>> rule.is_um_c_(2)
        nan
    """
    pass



# Voting Rule
from svvamp.rules.rule_approval import RuleApproval
from svvamp.rules.rule_baldwin import RuleBaldwin
from svvamp.rules.rule_black import RuleBlack
from svvamp.rules.rule_borda import RuleBorda
from svvamp.rules.rule_bucklin import RuleBucklin
from svvamp.rules.rule_condorcet_sum_defeats import RuleCondorcetSumDefeats
from svvamp.rules.rule_coombs import RuleCoombs
from svvamp.rules.rule_copeland import RuleCopeland
from svvamp.rules.rule_irv_duels import RuleIRVDuels
from svvamp.rules.rule_iterated_bucklin import RuleIteratedBucklin
from svvamp.rules.rule_kemeny import RuleKemeny
from svvamp.rules.rule_kim_roush import RuleKimRoush
from svvamp.rules.rule_majority_judgment import RuleMajorityJudgment
from svvamp.rules.rule_maximin import RuleMaximin
from svvamp.rules.rule_nanson import RuleNanson
from svvamp.rules.rule_plurality import RulePlurality
from svvamp.rules.rule_range_voting import RuleRangeVoting
from svvamp.rules.rule_ranked_pairs import RuleRankedPairs
from svvamp.rules.rule_schulze import RuleSchulze
from svvamp.rules.rule_split_cycle import RuleSplitCycle
from svvamp.rules.rule_star import RuleSTAR
from svvamp.rules.rule_two_round import RuleTwoRound
from svvamp.rules.rule_veto import RuleVeto

# Voting Rule: IRV Family
from svvamp.rules.rule_exhaustive_ballot import RuleExhaustiveBallot
from svvamp.rules.rule_irv import RuleIRV
from svvamp.rules.rule_condorcet_abs_irv import RuleCondorcetAbsIRV
from svvamp.rules.rule_condorcet_vtb_irv import RuleCondorcetVtbIRV
from svvamp.rules.rule_icrv import RuleICRV
from svvamp.rules.rule_irv_average import RuleIRVAverage
from svvamp.rules.rule_smith_irv import RuleSmithIRV
from svvamp.rules.rule_tideman import RuleTideman
from svvamp.rules.rule_woodall import RuleWoodall

ALL_RULE_CLASSES = [
    RuleApproval,
    RuleBaldwin, RuleBlack, RuleBorda, RuleBucklin,
    RuleCondorcetSumDefeats, RuleCoombs, RuleCopeland,
    RuleIRVDuels, RuleIteratedBucklin,
    RuleKemeny, RuleKimRoush,
    RuleMajorityJudgment, RuleMaximin,
    RuleNanson,
    RulePlurality,
    RuleRangeVoting, RuleRankedPairs,
    RuleSchulze, RuleSplitCycle, RuleSTAR,
    RuleTwoRound,
    RuleVeto,
    RuleExhaustiveBallot, RuleIRV, RuleCondorcetAbsIRV, RuleCondorcetVtbIRV,
    RuleICRV, RuleIRVAverage, RuleSmithIRV, RuleTideman, RuleWoodall,
]



# -*- coding: utf-8 -*-
"""
Created on nov. 28, 2018, 16:57
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import pandas as pd

from svvamp.utils.pref_lib import preflib_to_preferences_ut
from svvamp.utils.cast_vote_records import cvr_to_preferences_ut
from svvamp.preferences.profile import Profile


class ProfileFromFile(Profile):

    def __init__(self, file_name, sort_candidates=False, sort_voters=False):
        """Profile from a file.

        Parameters
        ----------
        file_name : str
            The name of the file.
        sort_candidates : bool
            If True, then candidates are sorted, first by number of victories (in the matrix of victories), then
            by Borda score.
        sort_voters : bool
            Whether the voters should be sorted. Cf. :class:`Profile`.

        Notes
        -----
        * If the file name ends with '.cvr.csv' (cvr = Cast Vote Records): first row gives the column headers. The
          last headers are 'rank1', 'rank2', etc. Each row represents a voter. In the column 'rank1', the name of
          her most-liked candidate, etc.
        * If the file name ends with '.t.csv' (t = transposed format): simple table of utilities with candidates
          declared in the first column and voters declared in the first row.
        * If the file name ends with '.csv' (but not '.t.csv'), candidates must be declared in the first row and voters
          in the first column.
        * Otherwise, the file is considered as a PrefLib file.

        For CVR and Preflib files, since the information is ordinal only, ``preferences_ut[v, c]`` is set to the Borda
        score (with no vtb) minus ``(C - 1) / 2``. This way, utilities are between ``- (C - 1) / 2`` and
        ``(C - 1) / 2``.
        """
        if file_name[-4:] == '.csv':
            if len(file_name) >= 8 and file_name[-8:] == '.cvr.csv':
                preferences_ut, labels_candidates = cvr_to_preferences_ut(file_name)
            else:
                df = pd.read_csv(filepath_or_buffer=file_name, sep=';', index_col=0)
                if file_name[-6:-4] == '.t':
                    preferences_ut = df.transpose().values
                    labels_candidates = df.index.values.astype(np.str)
                else:
                    preferences_ut = df.values
                    labels_candidates = df.columns.values.astype(np.str)
        else:
            preferences_ut, labels_candidates = preflib_to_preferences_ut(file_name)
        if sort_candidates:
            pop_temp = Profile(preferences_ut, sort_voters=False)
            nb_victories_temp = np.sum(pop_temp.matrix_victories_ut_rel, 1)
            scores_temp = nb_victories_temp + pop_temp.borda_score_c_ut / pop_temp.n_c / pop_temp.n_v
            candidates_best_to_worst = np.argsort(- scores_temp, kind='mergesort')
            preferences_ut = preferences_ut[:, candidates_best_to_worst]
            labels_candidates = np.array(labels_candidates)[candidates_best_to_worst]
        super().__init__(preferences_ut=preferences_ut,
                         log_creation=['From file', 'File name', file_name],
                         labels_candidates=labels_candidates, sort_voters=sort_voters)



# -*- coding: utf-8 -*-
"""
Created on jul. 21, 2021, 21:16
Copyright François Durand 2014-2021
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import inspect
import copy
from svvamp.meta.study_rule_criteria import StudyRuleCriteria
from svvamp.rules.all_rule_classes import ALL_RULE_CLASSES
from svvamp.rules.rule_approval import RuleApproval
from svvamp.rules.rule_plurality import RulePlurality
from svvamp.rules.rule_ranked_pairs import RuleRankedPairs
from svvamp.rules.rule_schulze import RuleSchulze
from svvamp.rules.rule_exhaustive_ballot import RuleExhaustiveBallot
from svvamp.rules.rule_irv import RuleIRV
from svvamp.rules.rule_condorcet_vtb_irv import RuleCondorcetVtbIRV
from svvamp.rules.rule_icrv import RuleICRV
from svvamp.utils.misc import indent


class VotingRuleTasks(list):
    """A set of tasks for the simulator, i.e. which voting rules with which options and which criteria about them.

    A "task" consists of: a voting rule, a dictionary of options, and a :class:`StudyRuleCriteria` object. A
    `VotingRuleTasks` is essentially a list of tasks.

    In the following, what we call a "voting system" consists of a voting rule and a dictionary of options (i.e.
    like a task, but with no :class:`StudyRuleCriteria` object specified).

    Parameters
    ----------
    voting_systems: List
        A list of pairs (rule_class, options), where `rule_class` is a subclass of :class:`Rule` and `options` is a
        dictionary of options. Instead of (rule_class, options), one may provide a singleton `rule_class`
        (without parentheses), which is equivalent to (rule_class, {}).
    study_rule_criteria: StudyRuleCriteria
        A :class:`StudyRuleCriteria` object.
    detailed_tasks: object
        Either a list of triples (rule_class, options, study_rule_criteria), or a VotingRuleTasks object.

    Examples
    --------
    Basically, there are two non-exclusive ways to define a VotingRuleTasks object:

        * Exhaustively with detailed_tasks.
        * As a Cartesian product by voting_systems and study_rule_criteria.

        >>> voting_rule_tasks1 = VotingRuleTasks(
        ...     voting_systems=[
        ...         RulePlurality,
        ...         RuleExhaustiveBallot,
        ...         (RuleIRV, {'cm_option': 'fast'}),
        ...         (RuleApproval, {'approval_threshold': 10}),
        ...     ],
        ...     study_rule_criteria=StudyRuleCriteria(
        ...         manipulation_criteria=['is_cm_', 'is_tm_', 'is_um_'],
        ...     ),
        ...     detailed_tasks=[
        ...         (
        ...             RuleIRV,
        ...             {'cm_option': 'exact'},
        ...             StudyRuleCriteria(
        ...                 manipulation_criteria=['is_cm_'],
        ...                 manipulation_only=True
        ...             )
        ...         ),
        ...         (
        ...             RuleApproval,
        ...             {'approval_threshold': 5},
        ...             StudyRuleCriteria(
        ...                 manipulation_criteria=['is_cm_', 'is_tm_', 'is_um_']
        ...             )
        ...         ),
        ...     ]
        ... )

        >>> voting_rule_tasks2 = VotingRuleTasks(
        ...     voting_systems=[RuleIRV, RuleExhaustiveBallot],
        ...     study_rule_criteria=StudyRuleCriteria(),
        ...     detailed_tasks=VotingRuleTasks(
        ...         voting_systems=[
        ...             (RuleIRV, {'cm_option': 'exact'}),
        ...             (RuleExhaustiveBallot, {'cm_option': 'exact'})
        ...         ],
        ...         study_rule_criteria=StudyRuleCriteria(
        ...             manipulation_criteria=['is_cm_'],
        ...             manipulation_only=True
        ...         )
        ...     )
        ... )

    In examples 1 and 2 above, all elements of `voting_systems` will be
    assigned the same set of criteria, the argument `study_rule_criteria`. The
    argument `detailed_tasks` allows to give different criteria for voting
    systems needing a specific treatment.

        * When voting_systems is None (default), it is normally set to [].
          But if detailed_tasks is empty, or study_rule_criteria is not empty,
          then voting_systems is set to a quite extensive list of
          voting systems.
        * When study_rule_criteria is None (default), it is normally set to [].
          But if detailed_tasks is empty, or voting_systems is not empty,
          then study_rule_criteria is set to StudyRuleCriteria(), which leads to study
          an quite extensive list of criteria.

    Study :class:`RuleSchulze` and :class:`Ranked Pairs` with a quite extensive list of criteria:

        >>> voting_rule_tasks = VotingRuleTasks(
        ...     voting_systems=[RuleSchulze, RuleRankedPairs]
        ... )

    Study a quite extensive list of voting systems, with CM being the only manipulation criterion:

        >>> voting_rule_tasks = VotingRuleTasks(
        ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'])
        ... )

    Study a quite extensive list of voting systems with a quite extensive list of criteria:

        >>> voting_rule_tasks = VotingRuleTasks()
    """

    def __init__(self, voting_systems=None, study_rule_criteria=None, detailed_tasks=None):
        # Dealing with default arguments...
        if detailed_tasks is None:
            detailed_tasks = []
        if voting_systems is None:
            if not detailed_tasks or study_rule_criteria is not None:
                voting_systems = ALL_RULE_CLASSES
            else:
                voting_systems = []
        if study_rule_criteria is None:
            study_rule_criteria = StudyRuleCriteria()
        # Here we go
        super().__init__([])
        for voting_system in voting_systems:
            if inspect.isclass(voting_system):
                self.append(
                    (voting_system, {}, copy.deepcopy(study_rule_criteria))
                )
            else:
                self.append(
                    (voting_system[0], voting_system[1], copy.deepcopy(study_rule_criteria))
                )
        self.extend(detailed_tasks)

    def __str__(self):
        """
            >>> voting_rule_tasks = VotingRuleTasks(detailed_tasks=[
            ...     (
            ...         RuleIRV,
            ...         {'cm_option': 'exact', 'um_option': 'exact'},
            ...         StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                           numerical_criteria=[])
            ...     ),
            ...     (
            ...         RuleCondorcetVtbIRV,
            ...         {'cm_option': 'very_slow', 'um_option': 'exact'},
            ...         StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                           numerical_criteria=[])
            ...     )
            ... ])
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleIRV
                options: {'cm_option': 'exact', 'um_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleCondorcetVtbIRV
                options: {'cm_option': 'very_slow', 'um_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        lines = []
        for (rule_class, options, study_rule_criteria) in self:
            lines.append(f'voting_system: {rule_class.__name__}')
            lines.append(f'options: {options}')
            lines.append(str(study_rule_criteria))
        return 'VotingRuleTasks with:\n' + indent('\n'.join(lines))

    def remove_rule(self, rule_class):
        """Remove a rule

        Removes all occurrences of `rule_class` from this `VotingRuleTasks` object.

        Parameters
        ----------
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
            >>> voting_rule_tasks = VotingRuleTasks(detailed_tasks=[
            ...     (
            ...         RuleIRV,
            ...         {'cm_option': 'fast', 'um_option': 'fast'},
            ...         StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                           numerical_criteria=[])
            ...     ),
            ...     (
            ...         RuleIRV,
            ...         {'cm_option': 'exact', 'um_option': 'exact'},
            ...         StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                           numerical_criteria=[])
            ...     ),
            ...     (
            ...         RuleCondorcetVtbIRV,
            ...         {'cm_option': 'very_slow', 'um_option': 'exact'},
            ...         StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                           numerical_criteria=[])
            ...     )
            ... ])
            >>> voting_rule_tasks.remove_rule(RuleIRV)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleCondorcetVtbIRV
                options: {'cm_option': 'very_slow', 'um_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        self[:] = [element for element in self if element[0] != rule_class]

    def append_task(self, rule_class, options=None, study_rule_criteria=None):
        """Append a task to the list.

        Parameters
        ----------
        rule_class : class
            Subclass of :class:`Rule`.
        options : dict
            A dictionary of options. If None (default), then options will be an empty dictionary, which amounts to
            using the default options for this voting rule.
        study_rule_criteria : StudyRuleCriteria
            If None (default), then use the default StudyRuleCriteria() (with a quite extensive list of criteria).

        Examples
        --------
            >>> voting_rule_tasks = VotingRuleTasks(detailed_tasks=[
            ...     (
            ...         RuleIRV,
            ...         {'cm_option': 'fast'},
            ...         StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                           numerical_criteria=[])
            ...     ),
            ... ])
            >>> voting_rule_tasks.append_task(
            ...     RuleIRV,
            ...     {'cm_option': 'exact'},
            ...     StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                       numerical_criteria=[])
            ... )
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleIRV
                options: {'cm_option': 'fast'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        if options is None:
            options = {}
        if study_rule_criteria is None:
            study_rule_criteria = StudyRuleCriteria()
        self.append((rule_class, options, copy.deepcopy(study_rule_criteria)))

    def extend(self, iterable):
        """Extend a `VotingRuleTasks` object.

        Note that the usual command `extend` can be used to concatenate two `VotingRuleTasks` objects.

        Examples
        --------
            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks_2 = VotingRuleTasks(
            ...     voting_systems=[RuleICRV, RuleCondorcetVtbIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_um_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.extend(voting_rule_tasks_2)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleICRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_um_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleCondorcetVtbIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_um_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        super().extend(iterable)

    def remove_criterion(self, criterion, rule_class=None):
        """Remove a criterion for one or all voting rules.

        Parameters
        ----------
        criterion : str
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Removes all occurrences of `criterion` for all occurrences of `rule_class`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV, (RuleIRV, {'cm_option': 'exact'})],
            ...     study_rule_criteria=StudyRuleCriteria(
            ...         manipulation_criteria=['is_cm_', 'is_um_'],
            ...         manipulation_only=True,
            ...         numerical_criteria=[]
            ...     )
            ... )
            >>> voting_rule_tasks.remove_criterion(criterion='is_um_', rule_class=RuleIRV)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                        is_um_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None

        If `rule_class` is None (default), then `criterion` is removed for all voting rules:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(
            ...         manipulation_criteria=['is_cm_', 'is_um_'],
            ...         manipulation_only=True,
            ...         numerical_criteria=[]
            ...     )
            ... )
            >>> voting_rule_tasks.remove_criterion('is_um_')
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        for (rc, options, study_rule_criteria) in self:
            if rule_class is None or rule_class == rc:
                study_rule_criteria.remove(criterion)

    def append_manipulation_criterion(self, criterion, rule_class=None):
        """Add a manipulation criterion for one or all voting rules.

        Parameters
        ----------
        criterion : str
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Add `criterion` for all occurrences of `rule_class`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV, (RuleIRV, {'cm_option': 'exact'})],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_manipulation_criterion(criterion='is_um_', rule_class=RuleIRV)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                        is_um_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                        is_um_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None

        If `rule_class` is None (default), then add `criterion` for all voting rules:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_manipulation_criterion('is_um_')
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                        is_um_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                        is_um_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        for (rc, options, study_rule_criteria) in self:
            if rule_class is None or rule_class == rc:
                study_rule_criteria.append_manipulation_criterion(criterion)

    def append_manipulation_criterion_c(self, criterion, rule_class=None):
        """Add a manipulation criterion (candidates) for one or all voting rules.

        Parameters
        ----------
        criterion : str
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Add `criterion` for all occurrences of `rule_class`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV, (RuleIRV, {'cm_option': 'exact'})],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_manipulation_criterion_c(criterion='candidates_cm_', rule_class=RuleIRV)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c:
                        candidates_cm_
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c:
                        candidates_cm_
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None

        If `rule_class` is None (default), then add `criterion` for all voting rules:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_manipulation_criterion_c('candidates_cm_')
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c:
                        candidates_cm_
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c:
                        candidates_cm_
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        for (rc, options, study_rule_criteria) in self:
            if rule_class is None or rule_class == rc:
                study_rule_criteria.append_manipulation_criterion_c(criterion)

    def append_result_criterion(self, criterion, rule_class=None):
        """Add a result criterion for one or all voting rules

        Parameters
        ----------
        criterion : str
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Add `criterion` for all occurrences of `rule_class`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV, (RuleIRV, {'cm_option': 'exact'})],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_result_criterion(criterion='w_is_condorcet_admissible_', rule_class=RuleIRV)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria:
                        w_is_condorcet_admissible_
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria:
                        w_is_condorcet_admissible_
                    utility_criteria: None
                    numerical_criteria: None

        If `rule_class` is None (default), then add `criterion` for all voting rules:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_result_criterion(criterion='w_is_condorcet_admissible_')
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria:
                        w_is_condorcet_admissible_
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria:
                        w_is_condorcet_admissible_
                    utility_criteria: None
                    numerical_criteria: None
        """
        for (rc, options, study_rule_criteria) in self:
            if rule_class is None or rule_class == rc:
                study_rule_criteria.append_result_criterion(criterion)

    def append_utility_criterion(self, criterion, func, name, rule_class=None):
        """Add a utility criterion for one or all voting rules

        Parameters
        ----------
        criterion : str
            The numerical data to be processed.
        func : callable
            An aggregating function.
        name : str
            The name chosen for this pair criterion / aggregation function. Cf.
            :meth:`StudyRuleCriteria.append_utility_criterion`.
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Add `criterion` for all occurrences of `rule_class`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV, (RuleIRV, {'cm_option': 'exact'})],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_utility_criterion(
            ...     criterion='total_utility_w_', func=np.sum, name='total_u_sum', rule_class=RuleIRV)
            >>> print(voting_rule_tasks)  # doctest: +ELLIPSIS
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria:
                        ('total_utility_w_', <function sum at ...>, 'total_u_sum')
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria:
                        ('total_utility_w_', <function sum at ...>, 'total_u_sum')
                    numerical_criteria: None

        If `rule_class` is None (default), then add `criterion` for all voting rules:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_utility_criterion(
            ...     criterion='total_utility_w_', func=np.sum, name='total_u_sum')
            >>> print(voting_rule_tasks)  # doctest: +ELLIPSIS
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria:
                        ('total_utility_w_', <function sum at ...>, 'total_u_sum')
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria:
                        ('total_utility_w_', <function sum at ...>, 'total_u_sum')
                    numerical_criteria: None
        """
        for (rc, options, study_rule_criteria) in self:
            if rule_class is None or rule_class == rc:
                study_rule_criteria.append_utility_criterion(criterion, func, name)

    def append_numerical_criterion(self, criterion, rule_class=None):
        """Add a numerical criterion for one or all voting rules

        Parameters
        ----------
        criterion : str
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Add `criterion` for all occurrences of `rule_class`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV, (RuleIRV, {'cm_option': 'exact'})],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_numerical_criterion(criterion='nb_candidates_cm_', rule_class=RuleIRV)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria:
                        nb_candidates_cm_
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria:
                        nb_candidates_cm_

        If `rule_class` is None (default), then add `criterion` for all voting rules:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.append_numerical_criterion(criterion='nb_candidates_cm_')
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria:
                        nb_candidates_cm_
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria:
                        nb_candidates_cm_
        """
        for (rc, options, study_rule_criteria) in self:
            if rule_class is None or rule_class == rc:
                study_rule_criteria.append_numerical_criterion(criterion)

    def remove_option(self, option, rule_class=None):
        """Remove an option and its value for one or all voting rules

        Parameters
        ----------
        option : str
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Remove `option` for all occurrences of `rule_class`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[
            ...         (RuleExhaustiveBallot, {'cm_option': 'exact'}),
            ...         (RuleIRV, {'cm_option': 'exact'})
            ...     ],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.remove_option(option='cm_option', rule_class=RuleIRV)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None

        This means that from now on, `RuleIRV` will use its default value for `cm_option`.

        If `rule_class` is None (default), then remove the option for all voting rules:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[
            ...         (RuleExhaustiveBallot, {'cm_option': 'exact'}),
            ...         (RuleIRV, {'cm_option': 'exact'})
            ...     ],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.remove_option(option='cm_option')
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        for (rc, options, study_rule_criteria) in self:
            if rule_class is None or rule_class == rc:
                if option in options:
                    del(options[option])

    def set_option(self, option, value, rule_class=None):
        """Set an option for one or all voting rules

        Parameters
        ----------
        option : str
        value : object
            New value for the option.
        rule_class : class
            Subclass of :class:`Rule`.

        Examples
        --------
        Set `option` to `value` for all occurrences of `rule_class`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.set_option('cm_option', 'exact', rule_class=RuleIRV)
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None

        If `rule_class` is None (default), do this for all voting rules:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleExhaustiveBallot, RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['is_cm_'], manipulation_only=True,
            ...                                           numerical_criteria=[])
            ... )
            >>> voting_rule_tasks.set_option('cm_option', 'exact')
            >>> print(voting_rule_tasks)
            VotingRuleTasks with:
                voting_system: RuleExhaustiveBallot
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
                voting_system: RuleIRV
                options: {'cm_option': 'exact'}
                StudyRuleCriteria with:
                    manipulation_criteria:
                        is_cm_
                    manipulation_criteria_c: None
                    result_criteria: None
                    utility_criteria: None
                    numerical_criteria: None
        """
        for (rc, options, study_rule_criteria) in self:
            if rule_class is None or rule_class == rc:
                # print(rule_class)
                # print('Change option to '+ str(value))
                options[option] = value

    def check_sanity(self):
        """Check sanity of the object

        Preform some basic checks. It is recommended to use this method before launching big simulations.

        Examples
        --------
        Detect an illegal option:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[(RuleIRV, {'cm_option': 'unexpected_option'})],
            ... )
            >>> voting_rule_tasks.check_sanity()
            Traceback (most recent call last):
            ValueError: 'cm_option' = 'unexpected_option' is not allowed in RuleIRV.

        Detect an illegal :class:`StudyRuleCriteria`:

            >>> voting_rule_tasks = VotingRuleTasks(
            ...     voting_systems=[RuleIRV],
            ...     study_rule_criteria=StudyRuleCriteria(manipulation_criteria=['unexpected_criterion'])
            ... )
            >>> voting_rule_tasks.check_sanity()
            Traceback (most recent call last):
            ValueError: Criterion 'unexpected_criterion' is unknown for RuleIRV.
        """
        for (rule_class, options, study_rule_criteria) in self:
            # print(rule_class)
            for option, value in options.items():
                rule_class.check_option_allowed(option, value)
            study_rule_criteria.check_sanity(rule_class)
        print('VotingRuleTasks: Sanity check was successful.')



from svvamp import RuleSchulze


def test_um_option_setter():
    """
        >>> rule = RuleSchulze()
        >>> rule.um_option = 'fast'
        >>> rule.um_option
        'fast'
        >>> rule.um_option = 'exact'
        >>> rule.um_option
        'exact'
        >>> rule.um_option = 'foo'
        Traceback (most recent call last):
        ValueError: Unknown value for um_option: foo
    """
    pass


def test_cm_option_setter():
    """
        >>> rule = RuleSchulze()
        >>> rule.cm_option = 'fast'
        >>> rule.cm_option
        'fast'
        >>> rule.cm_option = 'exact'
        >>> rule.cm_option
        'exact'
        >>> rule.cm_option = 'foo'
        Traceback (most recent call last):
        ValueError: Unknown value for cm_option: foo
    """
    pass



# -*- coding: utf-8 -*-
"""
Created on 4 dec. 2018, 16:16
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleCoombs(Rule):
    """Coombs method.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleCoombs()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [[-1. -2. -2.]
         [-2. -3. nan]]
        candidates_by_scores_best_to_worst
        [0, 1, 2]
        scores_best_to_worst
        [[-1. -2. -2.]
         [-2. -3. nan]]
        w = 0
        score_w = [-1. -2.]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (False)     ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (False)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (False) ==> maj_fav_c_rk (False)       ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (False) ==> majority_favorite_c_ut (False)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = fast
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = fast
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = fast, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 0. 0.]
        sufficient_coalition_size_cm =
        [0. 6. 4.]

    Notes
    -----
    The candidate who is ranked last by most voters is eliminated. Then we iterate. Ties are broken in favor of
    lower-index candidates: in case of a tie, the tied candidate with highest index is eliminated.

    * :meth:`is_cm_`:

        * :attr:`cm_option` = ``'fast'``: Polynomial heuristic. Can prove CM but unable to decide non-CM (except in
          rare obvious cases).
        * :attr:`cm_option` = ``'exact'``: Non-polynomial (:math:`n_c !`).

    * :meth:`is_icm_`: Exact in polynomial time.
    * :meth:`is_im_`:

        * :attr:`im_option` = ``'fast'``: Polynomial heuristic. Can prove IM but unable to decide non-IM (except in
          rare obvious cases).
        * :attr:`im_option` = ``'exact'``: Non-polynomial (:math:`n_c !`).

    * :meth:`is_iia`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: For this voting system, UM and CM are equivalent. For this reason, :attr:`um_option` and
      :attr:`cm_option` are linked to each other: modifying one modifies the other accordingly.

    References
    ----------
    'On The Complexity of Manipulating Elections', Tom Coleman and Vanessa Teague, 2007.
    """

    full_name = 'Coombs'
    abbreviation = 'Coo'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'im_option': {'allowed': ['fast', 'exact'], 'default': 'fast'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'um_option': {'allowed': ['fast', 'exact'], 'default': 'fast'},
        'icm_option': {'allowed': ['exact'], 'default': 'exact'},
        'cm_option': {'allowed': ['fast', 'exact'], 'default': 'fast'},
    })

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_um=False, precheck_icm=False,
            log_identity="COOMBS", **kwargs
        )

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        scores = np.zeros((self.profile_.n_c - 1, self.profile_.n_c))
        is_eliminated = np.zeros(self.profile_.n_c, dtype=np.bool)
        worst_to_best = []
        one_v_might_be_pivotal = False
        # preferences_borda_temp : we will put ``n_c + 1`` for eliminated candidates
        preferences_borda_temp = np.copy(self.profile_.preferences_borda_rk)
        for r in range(self.profile_.n_c - 1):
            scores[r, :] = - np.bincount(np.argmin(preferences_borda_temp, 1), minlength=self.profile_.n_c)
            scores[r, is_eliminated] = np.nan
            loser = np.where(scores[r, :] == np.nanmin(scores[r, :]))[0][-1]  # Tie-breaking: the last index
            is_eliminated[loser] = True
            worst_to_best.append(loser)
            preferences_borda_temp[:, loser] = self.profile_.n_c + 1
            # Dealing with the possibility of pivotality
            # ``margins_r[c] = 0`` : ``c`` is eliminated (i.e. ``c == loser``).
            # ``margins_r[c] = x`` means that with ``x`` vetos more, ``c`` would be eliminated instead of ``loser``.
            worst_score = scores[r, loser]
            margins_r = scores[r, :] - worst_score + (np.array(range(self.profile_.n_c)) < loser)
            self.mylogv("margins_r =", margins_r, 3)
            self.mylogv("np.max(margins_r) =", np.nanmax(margins_r), 3)
            if np.nanmax(margins_r) <= 2:
                one_v_might_be_pivotal = True
        w = np.argmin(is_eliminated)
        worst_to_best.append(w)
        candidates_by_scores_best_to_worst = worst_to_best[::-1]
        return {'scores': scores, 'one_v_might_be_pivotal': one_v_might_be_pivotal, 'w': w,
                'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst}

    @cached_property
    def scores_(self):
        """2d array of integers. ``scores[r, c]`` is minus the number of voters who vote against candidate ``c`` at
        elimination round ``r``.
        """
        return self._count_ballots_['scores']

    @cached_property
    def w_(self):
        """Integer (winning candidate)."""
        return self._count_ballots_['w']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. Candidates are sorted according to their order of elimination. By definition /
        convention, ``candidates_by_scores_best_to_worst_[0]`` = :attr:`w_`.
        """
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    @cached_property
    def _one_v_might_be_pivotal_(self):
        return self._count_ballots_['one_v_might_be_pivotal']

    @cached_property
    def v_might_im_for_c_(self):
        return np.full((self.profile_.n_v, self.profile_.n_c), self._one_v_might_be_pivotal_)

    @cached_property
    def meets_ignmc_c_ctb(self):
        return True

    # %% Setting the options

    @property
    def um_option(self):
        return self._um_option

    @um_option.setter
    def um_option(self, value):
        if self._um_option == value:
            return
        if value in self.options_parameters['um_option']['allowed']:
            self.mylogv("Setting um_option =", value, 1)
            self._cm_option = value
            self._um_option = value
            self.delete_cache(contains='_um_', suffix='_')
            self.delete_cache(contains='_cm_', suffix='_')
        else:
            raise ValueError("Unknown value for um_option: " + format(value))

    @property
    def cm_option(self):
        return self._cm_option

    @cm_option.setter
    def cm_option(self, value):
        if self._cm_option == value:
            return
        if value in self.options_parameters['cm_option']['allowed']:
            self.mylogv("Setting cm_option =", value, 1)
            self._cm_option = value
            self._um_option = value
            self.delete_cache(contains='_um_', suffix='_')
            self.delete_cache(contains='_cm_', suffix='_')
        else:
            raise ValueError("Unknown value for cm_option: " + format(value))

    # %% Manipulation: general routine

    def _cm_aux_fast(self, c, n_max, preferences_borda_s):
        """Fast algorithm used for IM and CM (which is equivalent to UM)

        Parameters
        ----------
        c : int
            Candidate for which we want to manipulate.
        n_max : int or inf
            Maximum number of manipulators allowed.

            * IM --> put 1.
            * CM, complete and exact --> put the current value of ``sufficient_coalition_size[c] - 1`` (we want to find
                the best value for ``sufficient_coalition_size[c]``, even if it exceeds the number of manipulators).
            * CM, otherwise --> put the number of manipulators.

        preferences_borda_s : list of list
            Preferences of the sincere voters (in Borda format with vtb).

        Returns
        -------
        int or inf
            ``n_manip_fast``. If a manipulation is found with n_max manipulators or less, then a sufficient number
            of manipulators is returned. Otherwise, it is +inf.

        Examples
        --------
            >>> profile = Profile(preferences_rk=[
            ...     [1, 2, 0, 3],
            ...     [1, 3, 2, 0],
            ...     [2, 0, 3, 1],
            ...     [2, 1, 3, 0],
            ...     [3, 2, 0, 1],
            ... ])
            >>> rule = RuleCoombs()(profile)
            >>> rule.candidates_cm_
            array([0., 1., 0., 1.])
        """
        # Each step of the loop:
        # ^^^^^^^^^^^^^^^^^^^^^^
        # We start at the end of round ``r``, with ``candidates[is_candidate_alive_end_r]``. We find the ``d`` such
        # that, with ``d`` in addition at the beginning, we need the fewest manipulators to eliminate ``d`` (and reach
        # the final situation we wanted).
        # Initialization
        # ^^^^^^^^^^^^^^
        # End of last round: only ``c`` is alive.
        nb_manipulators_used = 0
        is_candidate_alive_end_r = np.zeros(self.profile_.n_c, dtype=np.bool)
        is_candidate_alive_end_r[c] = True
        for r in range(self.profile_.n_c - 2, -1, -1):
            self.mylogv("cm_aux_fast: Round r =", r, 3)
            least_nb_manipulators_for_r = np.inf
            # We look for a candidate with which round r was easy
            best_d = None
            for d in range(self.profile_.n_c):
                if is_candidate_alive_end_r[d]:
                    continue
                self.mylogv("cm_aux_fast: Try to add d =", d, 3)
                is_candidate_alive_begin_r = np.copy(is_candidate_alive_end_r)
                is_candidate_alive_begin_r[d] = True
                scores_s = np.zeros(self.profile_.n_c)
                scores_s[is_candidate_alive_begin_r] = - np.sum(np.equal(
                    preferences_borda_s[:, is_candidate_alive_begin_r],
                    np.min(preferences_borda_s[:, is_candidate_alive_begin_r], 1)[:, np.newaxis]
                ), 0)
                # self.mylogv("cm_AUX: scores_s =", scores_s, 3)
                normal_loser = np.where(scores_s == np.min(scores_s))[0][-1]  # Tie-breaking
                nb_manip_d = max(0, scores_s[d] - scores_s[normal_loser] + (d < normal_loser))
                self.mylogv("cm_aux_fast: nb_manip_d =", nb_manip_d, 3)
                if nb_manip_d < least_nb_manipulators_for_r:
                    least_nb_manipulators_for_r = nb_manip_d
                    best_d = d
            self.mylogv("cm_aux_fast: best_d =", best_d, 3)
            # Update variables for previous round
            nb_manipulators_used = max(nb_manipulators_used, least_nb_manipulators_for_r)
            if nb_manipulators_used > n_max:
                self.mylogv("cm_aux_fast: Conclusion: nb_manipulators_used =", np.inf, 3)
                return np.inf
            is_candidate_alive_end_r[best_d] = True

        self.mylogv("cm_aux_fast: Conclusion: nb_manipulators_used =", nb_manipulators_used, 3)
        return nb_manipulators_used

    def _cm_aux_exact(self, c, n_max, preferences_borda_s):
        """Exact algorithm used for IM and CM (which is equivalent to UM)

        Parameters
        ----------
        c : int
            Candidate for which we want to manipulate.
        n_max : int
            Maximum number of manipulators allowed.

            * IM --> put 1.
            * CM, complete and exact --> put the current value of ``sufficient_coalition_size[c] - 1`` (we want to find
                the best value for ``sufficient_coalition_size[c]``, even if it exceeds the number of manipulators).
            * CM, otherwise --> put the number of manipulators.

        preferences_borda_s : list of list
            Preferences of the sincere voters (in Borda format).

        Returns
        -------
        int or inf
            ``n_manip_exact``. If a manipulation is found with ``n_max`` manipulators or less, the minimal sufficient
            number of manipulators is returned. Otherwise, it is +inf.

        Examples
        --------
            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1, 3],
            ...     [0, 2, 3, 1],
            ...     [0, 2, 3, 1],
            ...     [3, 0, 2, 1],
            ...     [3, 1, 2, 0],
            ... ])
            >>> rule = RuleCoombs(cm_option='exact')(profile)
            >>> rule.necessary_coalition_size_cm_
            array([0., 4., 4., 3.])
        """
        # Explore subsets that are reachable with less than the upper bound.
        # ``situations_end_r`` is a dictionary.
        # * keys: ``is_candidate_alive_end_r``, tuple of booleans.
        # * values: ``nb_manip_used_after_r``, number of manipulators used after round ``d`` to go from situation
        #   ``is_candidate_alive_end_r`` to the singleton ``c``.
        # Example: ``(1,1,0,1,0):3`` means that if we have candidates 0, 1 and 3, we need 3 manipulators to make ``c``
        # win at the end.
        situations_end_r = {tuple(np.array(range(self.profile_.n_c)) == c): 0}
        situations_begin_r = {}
        for r in range(self.profile_.n_c - 2, -1, -1):
            self.mylogv("cm_aux_exact: Round r =", r, 3)
            situations_begin_r = {}
            for is_candidate_alive_end_r, nb_manip_used_after_r in situations_end_r.items():
                self.mylogv("cm_aux_exact: is_candidate_alive_end_r =", is_candidate_alive_end_r, 3)
                self.mylogv("cm_aux_exact: nb_manip_used_after_r =", nb_manip_used_after_r, 3)
                for d in range(self.profile_.n_c):
                    if is_candidate_alive_end_r[d]:
                        continue
                    self.mylogv("cm_aux_exact: d =", d, 3)
                    is_candidate_alive_begin_r = np.copy(is_candidate_alive_end_r)
                    is_candidate_alive_begin_r[d] = True
                    scores_s = np.zeros(self.profile_.n_c)
                    scores_s[is_candidate_alive_begin_r] = - np.sum(np.equal(
                        preferences_borda_s[:, is_candidate_alive_begin_r],
                        np.min(preferences_borda_s[:, is_candidate_alive_begin_r], 1)[:, np.newaxis]
                    ), 0)
                    # self.mylogv("scores_s =", scores_s, 3)
                    normal_loser = np.where(scores_s == np.nanmin(scores_s))[0][-1]  # Tie-breaking
                    nb_manip_r = max(scores_s[d] - scores_s[normal_loser] + (d < normal_loser), 0)
                    self.mylogv("cm_aux_exact: nb_manip_r =", nb_manip_r, 3)
                    nb_manip_r_and_after = max(nb_manip_r, nb_manip_used_after_r)
                    if nb_manip_r_and_after > n_max:
                        continue
                    if tuple(is_candidate_alive_begin_r) in situations_begin_r:
                        situations_begin_r[tuple(is_candidate_alive_begin_r)] = min(
                            situations_begin_r[tuple(is_candidate_alive_begin_r)],
                            nb_manip_r_and_after)
                    else:
                        situations_begin_r[tuple(is_candidate_alive_begin_r)] = nb_manip_r_and_after
            self.mylogv("cm_aux_exact: situations_begin_r =", situations_begin_r, 3)
            if len(situations_begin_r) == 0:
                self.mylog("cm_aux_exact: Manipulation is impossible with n_max manipulators.", 3)
                return np.inf  # By convention
            situations_end_r = situations_begin_r
        else:
            self.mylogv("cm_aux_exact: situations_begin_r:", situations_begin_r, 3)
            is_candidate_alive_begin, nb_manip_used = situations_begin_r.popitem()
            self.mylogv("cm_aux_exact: is_candidate_alive_begin:", is_candidate_alive_begin, 3)
            self.mylogv("cm_aux_exact: Conclusion: nb_manip_used_new =", nb_manip_used, 3)
        return nb_manip_used

    # %% Individual manipulation (IM)

    def _im_main_work_v_(self, v, c_is_wanted, nb_wanted_undecided, stop_if_true):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 3, 1],
            ...     [1, 0, 3, 2],
            ...     [1, 3, 2, 0],
            ...     [2, 1, 3, 0],
            ...     [3, 1, 0, 2],
            ... ])
            >>> rule = RuleCoombs()(profile)
            >>> rule.is_im_
            True

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCoombs()(profile)
            >>> rule.is_im_c_with_voters_(2)
            (nan, array([ 0.,  0.,  0., nan, nan]))

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCoombs(im_option='exact')(profile)
            >>> rule.is_im_
            False

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCoombs(im_option='exact')(profile)
            >>> rule.is_im_
            True

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleCoombs()(profile)
            >>> rule.v_im_for_c_
            array([[ 0.,  0.,  0.],
                   [ 0., nan,  0.],
                   [ 0., nan,  1.],
                   [ 0.,  0., nan],
                   [ 0.,  0., nan]])
        """
        preferences_borda_s = self.profile_.preferences_borda_rk[np.array(range(self.profile_.n_v)) != v, :]
        for c in range(self.profile_.n_c):
            if not c_is_wanted[c]:
                continue
            if not np.isneginf(self._v_im_for_c[v, c]):
                continue
            self.mylogv('IM: Candidate c =', c, 3)
            n_manip_fast = self._cm_aux_fast(c, n_max=1, preferences_borda_s=preferences_borda_s)
            self.mylogv("IM: Fast: n_manip_fast =", n_manip_fast, 3)
            if n_manip_fast <= 1:
                self._v_im_for_c[v, c] = True
                self._candidates_im[c] = True
                self._voters_im[v] = True
                self._is_im = True
                if stop_if_true:
                    return
                continue  # pragma: no cover - Not really "executed" because of Python optimization
            if self.im_option != 'exact':
                self._v_im_for_c[v, c] = np.nan
                continue
            n_manip_exact = self._cm_aux_exact(c, n_max=1, preferences_borda_s=preferences_borda_s)
            self.mylogv("IM: Exact: n_manip_exact =", n_manip_exact, 3)
            if n_manip_exact <= 1:
                self._v_im_for_c[v, c] = True
                self._candidates_im[c] = True
                self._voters_im[v] = True
                self._is_im = True
                if stop_if_true:
                    return
            else:
                self._v_im_for_c[v, c] = False

    # %% Trivial Manipulation (TM)

    # Use the generic methods from class Rule.

    # %% Unison manipulation (UM)

    @cached_property
    def is_um_(self):
        return self.is_cm_

    def is_um_c_(self, c):
        return self.is_cm_c_(c)

    @cached_property
    def candidates_um_(self):
        return self.candidates_cm_

    # %% Ignorant-Coalition Manipulation (ICM)

    # The voting system meets IgnMC_c_ctb: hence, general methods are exact.

    # %% Coalition Manipulation (CM)

    def _cm_main_work_c_(self, c, optimize_bounds):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1, 3],
            ...     [0, 2, 3, 1],
            ...     [0, 2, 3, 1],
            ...     [3, 0, 2, 1],
            ...     [3, 1, 2, 0],
            ... ])
            >>> rule = RuleCoombs(cm_option='exact')(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleCoombs(cm_option='exact')(profile)
            >>> rule.necessary_coalition_size_cm_
            array([0., 5., 4.])

            >>> profile = Profile(preferences_rk=[
            ...     [1, 3, 0, 2],
            ...     [2, 0, 3, 1],
            ...     [2, 1, 3, 0],
            ...     [3, 1, 0, 2],
            ...     [3, 2, 1, 0],
            ... ])
            >>> rule = RuleCoombs(cm_option='exact')(profile)
            >>> rule.candidates_cm_
            array([0., 1., 1., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [1, 2, 0, 3],
            ...     [1, 3, 2, 0],
            ...     [2, 0, 3, 1],
            ...     [2, 1, 3, 0],
            ...     [3, 2, 0, 1],
            ... ])
            >>> rule = RuleCoombs(cm_option='exact')(profile)
            >>> rule.candidates_cm_
            array([0., 1., 0., 1.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 0. ,  0. ],
            ...     [ 0.5, -0.5],
            ...     [-0.5, -1. ],
            ...     [-0.5,  0. ],
            ...     [ 0. ,  0.5],
            ...     [-0.5, -0.5],
            ... ], preferences_rk=[
            ...     [0, 1],
            ...     [0, 1],
            ...     [0, 1],
            ...     [1, 0],
            ...     [1, 0],
            ...     [1, 0],
            ... ])
            >>> rule = RuleCoombs(cm_option='exact')(profile)
            >>> rule.is_cm_c_(1)
            False
            >>> rule.sufficient_coalition_size_cm_
            array([0., 3.])
        """
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        exact = (self.cm_option == "exact")
        if optimize_bounds and exact:
            n_max = self._sufficient_coalition_size_cm[c] - 1
        else:
            n_max = n_m
        self.mylogv("CM: n_max =", n_max, 3)
        if not exact and self._necessary_coalition_size_cm[c] > n_max:  # pragma: no cover
            # TO DO: Investigate whether this case can actually happen.
            self._reached_uncovered_code()
            self.mylog("CM: Fast algorithm will not do better than what we already know", 3)
            return
        n_manip_fast = self._cm_aux_fast(
            c, n_max,
            preferences_borda_s=self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :])
        self.mylogv("CM: n_manip_fast =", n_manip_fast, 3)
        self._update_sufficient(self._sufficient_coalition_size_cm, c, n_manip_fast,
                                'CM: Update sufficient_coalition_size_cm =')
        if not exact:
            # With fast algo, we stop here anyway. It is not a "quick escape" (if we'd try again with
            # ``optimize_bound``, we would not try better).
            return

        # From this point, we have necessarily the 'exact' option.
        if self._sufficient_coalition_size_cm[c] == self._necessary_coalition_size_cm[c]:
            return
        if not optimize_bounds and n_m >= self._sufficient_coalition_size_cm[c]:
            # This is a quick escape: since we have the option 'exact', if we come back with ``optimize_bound``,
            # we will try to be more precise.
            return True
        # Either we're in complete mode (and might have succeeded), or we in incomplete mode (and we have failed).
        n_max_updated = min(n_manip_fast - 1, n_max)
        self.mylogv("CM: n_max_updated =", n_max_updated, 3)
        n_manip_exact = self._cm_aux_exact(
            c, n_max_updated,
            preferences_borda_s=self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :])
        self.mylogv("CM: n_manip_exact =", n_manip_exact)
        n_manip = min(n_manip_fast, n_manip_exact)
        self.mylogv("CM: n_manip =", n_manip, 3)
        self._update_sufficient(self._sufficient_coalition_size_cm, c, n_manip,
                                'CM: Update sufficient_coalition_size_cm =')
        # Update necessary coalition and return
        if optimize_bounds:
            self._necessary_coalition_size_cm[c] = self._sufficient_coalition_size_cm[c]
            return False
        else:
            if n_m >= n_manip:
                # We have optimized the size of the coalition.
                self._necessary_coalition_size_cm[c] = self._sufficient_coalition_size_cm[c]
                return False
            else:
                # We have explored everything with ``n_max = n_m`` but manipulation failed. However, we have not
                # optimized sufficient_size (which must be higher than ``n_m``), so it is a quick escape.
                self._update_necessary(self._necessary_coalition_size_cm, c, n_m + 1,
                                       'CM: Update necessary_coalition_size_cm = n_m + 1 =')
                return True
